/*
 * term.c
 * terminal handling routines for cancan
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>

#ifdef USETERMIO
#ifdef APOLLO
#include "/sys5.3/usr/include/sys/termio.h"
#else
/*
 * including both termio.h and termios.h might be an overkill, and gives
 * many warnings, but seems to be necessary at times. works anyway.
 */
/* necessary? when? */
/*#include <termio.h>*/
#include <termios.h>
#endif
#else /* not USETERMIO */
#include <sys/ioctl.h>
#endif

#ifdef BSD_386
#include <sys/ioctl_compat.h>
#define O_RAW RAW
#define O_ECHO ECHO
#define O_CBREAK CBREAK
#endif

#if defined(TCSETS) || defined(TCSETATTR)
#ifndef TCSETS		/* cc for HP-UX  SHOULD define this... */
#define TCSETS TCSETATTR
#define TCGETS TCGETATTR
#endif
typedef struct termios termiostruct;
#else
#define TCSETS TCSETA
#define TCGETS TCGETA
typedef struct termio termiostruct;
#endif

#include "cancan.h"
#include "edit.h"
#include "list.h"
#include "term.h"

int ioctl();

#ifdef VT100	/* hard-coded vt100 features if no termcap: */

char kpadstart[] = "", kpadend[] = "";
char begoln[] = "\r";
char clreoln[] = "\033[K", clreoscr[] = "\033[J";
char leftcur[] = "\033[D", rightcur[] = "\033[C", upcur[] = "\033[A";
char modebold[] = "\033[1m", modeinv[] = "\033[7m", modeuline[] = "\033[4m";
char modestandon[] = "", modestandoff[] = "", modenorm[] = "\033[m";

static int len_begoln = 1;
static int len_leftcur = 3, len_upcur = 3, gotocost = 8;
int metakey = 0, wrapglitch = 1;
#define insertfinish (0)

#else		/* termcap function declarations */

int tgetent();
int tgetnum();
int tgetflag();
char *tgetstr();
char *tgoto();

/* terminal escape sequences */
char kpadstart[CAPLEN], kpadend[CAPLEN];
char begoln[CAPLEN], clreoln[CAPLEN], clreoscr[CAPLEN];
char leftcur[CAPLEN], rightcur[CAPLEN], upcur[CAPLEN], curgoto[CAPLEN];
char delchar[CAPLEN], inschar[CAPLEN], insstop[CAPLEN];
	
/* attribute changers: */
char modebold[CAPLEN], modeinv[CAPLEN], modeuline[CAPLEN];
char modestandon[CAPLEN], modestandoff[CAPLEN], modenorm[CAPLEN];

static int len_begoln, len_clreoln;
static int len_leftcur, len_upcur, gotocost;
static int deletecost, insertcost, insertfinish;

int metakey;			/* terminal has meta key setting high bit */
int wrapglitch;			/* strange wrap behaviour (a la vt100) */

static int extract();

#endif /* VT100 */

#ifdef USETERMIO
static termiostruct ttybsave;
#else
static struct sgttyb ttybsave;
static struct tchars tcsave;
static struct ltchars ltcsave;
#endif

/*
 * Terminal handling routines:
 * These are one big mess of left-justified chicken scratches.
 */

/*
 * Save the original state, disable echo and special keys.
 */
void set_terminal()
{
#ifdef USETERMIO
    ioctl(0, TCGETS, &ttybsave);
#else /* not USETERMIO */
    ioctl(0, TIOCGETP, &ttybsave);
    ioctl(0, TIOCGETC, &tcsave);
    ioctl(0, TIOCGLTC, &ltcsave);
#endif
    set_spec_keys(0);
    fputs(kpadstart, stdout);
}

/*
 * Reset the terminal to its original state
 */
void reset_terminal()
{
#ifdef USETERMIO
    ioctl(0, TCSETS, &ttybsave);
#else
    ioctl(0, TIOCSETP, &ttybsave);
    ioctl(0, TIOCSETC, &tcsave);
    ioctl(0, TIOCSLTC, &ltcsave);
#endif
    fputs(kpadend, stdout);
}

/*
 * set raw/cooked mode depending on the current linemode
 * enable/disable special keys
 */
void set_spec_keys(k)
int k;
{
#ifdef USETERMIO
    termiostruct ttyb;
    ttyb = ttybsave;
    ttyb.c_oflag |= OPOST;
    if(linemode & LM_CHAR) ttyb.c_oflag &= ~OPOST;
    ttyb.c_lflag &= ~(ECHO | ISIG | ICANON | IEXTEN);
    if(k) ttyb.c_lflag |= ISIG;
    ioctl(0, TCSETS, &ttyb);
#else /* not USETERMIO */
    static struct tchars tc_off = { -1, -1, -1, -1, -1, -1 };
    static struct ltchars ltc_off = { -1, -1, -1, -1, -1, -1 };
    struct tchars *tc = &tc_off;
    struct ltchars *ltc = &ltc_off;
    struct sgttyb ttyb;
    ttyb = ttybsave;
    ttyb.sg_flags &= ~(O_RAW | O_ECHO);
    if(linemode & LM_CHAR) ttyb.sg_flags |= O_RAW;
    if(k) {
       tc = &tcsave;  ltc = &ltcsave;
    }
    ioctl(0, TIOCSETP, &ttyb);
    ioctl(0, TIOCSETC, tc);
    ioctl(0, TIOCSLTC, ltc);
#endif
}

/*
 * read termcap definitions
 */
void settermcap()
{
#ifndef VT100
    struct tc_init_node {
	char cap[4], *buf;
	int *len, critic;
    };
    static struct tc_init_node tc_init[] = {
	{ "cm", curgoto, 0, 1 },
	{ "ce", clreoln, &len_clreoln, 1 },
	{ "cd", clreoscr, 0, 1 },
	{ "nd", rightcur, 0, 1 },
	{ "le", leftcur, &len_leftcur, 0 },
	{ "up", upcur, &len_upcur, 0 },
	{ "cr", begoln, &len_begoln, 0 },
	{ "im", inschar, &insertcost, 0 },
	{ "ic", inschar, &insertcost, 0 },
	{ "ei", insstop, &insertcost, 0 },
	{ "dm", delchar, &deletecost, 0 },
	{ "dc", delchar, &deletecost, 0 },
	{ "ed", delchar, &deletecost, 0 },
	{ "me", modenorm, 0, 0 },
	{ "md", modebold, 0, 0 },
	{ "mr", modeinv, 0, 0 },
	{ "us", modeuline, 0, 0 },
	{ "so", modestandon, 0, 0 },
	{ "se", modestandoff, 0, 0 },
	{ "ks", kpadstart, 0, 0 },
	{ "ke", kpadend, 0, 0 },
	{ "", NULL, 0, 0 }
    };
    struct tc_init_node *np;
    char tcbuf[2048];		/* by convention, this is enough */
    char buf[CAPLEN];
    int i;
#endif    
#if !defined(VT100) || defined(TELNETBUG)
    char *term = getenv("TERM");
    if(!term) {
	printf("$TERM not set\n"); 
	exit(1);
    }
#endif
#ifdef VT100
    cols = 80; cols_1 = 79;
#ifdef LINES
    lines = LINES; lines_1 = LINES - 1;
#else
    lines = 24; lines_1 = 23;
#endif
    add_keynode("left", "\033[D", prev_char, NULL);
    add_keynode("right", "\033[C", next_char, NULL);
    add_keynode("up", "\033[A", prev_line, NULL);
    add_keynode("down", "\033[B", next_line, NULL);
#else
    switch(tgetent(tcbuf, term)) {
    case 1:
	break;
    case 0:
	fprintf(stderr,
	    "There is no entry for '%s' in the terminal data base.\n", term);
	fprintf(stderr,
	    "Please set your $TERM environment variable correctly.\n");
	exit(1);
    default:
	syserr("tgetent");
    }
    for(np = tc_init; np->cap[0]; np++)
	if((i = extract(np->cap, np->buf))) {
	    if(np->len) *np->len += i;
	} else if(np->critic) {
	    fprintf(stderr,
		    "Your terminal is not powerful enough, missing '%s'.\n",
		    np->cap);
	    exit(1);
	}
    if(!len_begoln)
	strcpy(begoln, "\r"), len_begoln = 1;
    if(!len_leftcur)
	strcpy(leftcur, "\b"), len_leftcur = 1;
    if(*buf = 0, extract("kl", buf))
	{ add_keynode("left", buf, prev_char, NULL);  *buf = 0; }
    if(extract("kr", buf))
	{ add_keynode("right", buf, next_char, NULL);  *buf = 0; }
    if(extract("ku", buf))
	{ add_keynode("up", buf, prev_line, NULL);  *buf = 0; }
    if(extract("kd", buf))
	add_keynode("down", buf, next_line, NULL);
    cols_1 = (cols = tgetnum("co")) - 1;
    lines_1 = (lines = tgetnum("li")) - 1;
    wrapglitch = tgetflag("xn");
    metakey = tgetflag("km");
    gotocost = strlen(tgoto(curgoto, cols_1, lines_1));
    insertfinish = gotocost + len_clreoln;
#endif
#ifdef TELNETBUG
    if(strncmp(term, "vt10", 4) == 0) {
	/* might be NCSA Telnet 2.2 for PC, which doesn't reset colours */
	sprintf(modenorm, "\033[;3%d;4%dm", DEFAULTFG, DEFAULTBG);
    }
#endif
}

/*
 * add the default keypad bindings to the list
 */
void default_keybindings()
{
    /*
     * Note: termcap doesn't have sequences for the numeric keypad, so we just
     * assume they are the same as for a vt100.
     */
    add_keynode("kp2", "\033Or", key_run_command, "s");
    add_keynode("kp3", "\033Os", key_run_command, "d");
    add_keynode("kp4", "\033Ot", key_run_command, "w");
    add_keynode("kp5", "\033Ou", key_run_command, "exits");
    add_keynode("kp6", "\033Ov", key_run_command, "e");
    add_keynode("kp8", "\033Ox", key_run_command, "n");
    add_keynode("kp9", "\033Oy", key_run_command, "u");
}
    
/*
 * initialize the key binding list
 */
void init_keybindings()
{
    struct b_init_node {
	char label[8], seq[4];
	void (*funct)();
    };
    static struct b_init_node b_init[] = {
	{ "",		"\n",		enter_line },
	{ "",		"\r",		enter_line },
	{ "bs",		"\b",		del_char_left },
	{ "del",	"\177",		del_char_left },
	{ "tab",	"\t",		complete_word },
	{ "C-a",	"\001",		beginning_of_line },
	{ "C-b",	"\002",		prev_char },
	{ "",		"\003",		intrsig },
	{ "C-d",	"\004",		del_char_right },
	{ "C-e",	"\005",		end_of_line },
	{ "C-f",	"\006",		next_char },
	{ "C-k",	"\013",		kill_to_eol },
	{ "C-l",	"\014",		redraw_input_line },
	{ "C-n",	"\016",		next_line },
	{ "C-p",	"\020",		prev_line },
	{ "C-t",	"\024",		transpose_chars },
	{ "C-y",	"\031",		enter_into_history },
	{ "C-z",	"\032",		suspend_cancan },
	{ "M-tab",	"\033\t",	complete_line },
	{ "M-b",	"\033b",	prev_word },
	{ "M-d",	"\033d",	del_word_right },
	{ "M-f",	"\033f",	next_word },
	{ "M-del",	"\033\177",	del_word_left },
	{ "",		"",		0	}
    };
    struct b_init_node *p = b_init;
    do {
	add_keynode(p->label, p->seq, p->funct, NULL);
    } while((++p)->seq[0]);
}

#ifndef VT100
/*
 * extract termcap 'cap' and strcat it to buf.
 * return the lenght of the extracted string.
 */
static int extract(cap, buf)
char *cap, *buf;
{
    static char *bp;
    char *d = buf + strlen(buf);
    char *s = tgetstr(cap, (bp = d, &bp));
    int len;
    if(!s) return (*bp = 0);
    /*
     * Remove the padding information. We assume that no terminals
     * need padding nowadays. At least it makes things much easier.
     */
    s += strspn(s, "0123456789*");
    for(len = 0; *s; *d++ = *s++, len++)
	if(*s == '$' && *(s + 1) == '<')
	    if(!(s = strchr(s, '>')) || !*++s) break;
    *d = 0;
    return len;
}
#endif /* VT100 */

/*
 * position the cursor using absolute coordinates
 */
void gotoxy(col, line)
int col, line;
{
#ifdef VT100
  printf("\033[%d;%dH", line + 1, col + 1);
#else
  fputs(tgoto(curgoto, col, line), stdout);
#endif
}

/*
 * optimized cursor movement
 * from (fromline, fromcol) to (toline, tocol)
 * if tocol > 0, (toline, tocol) must lie on editline.
 */
void optmove(fromline, fromcol, toline, tocol)
int fromline, fromcol, toline, tocol;
{
    static char buf[CAPLEN];
    char *cp = buf;
    int cost, i, dist;
    /* First, move vertically to the correct line, then horisontally
     * to the right column. If this turns out to be fewer characters
     * than a direct cursor positioning (gotoxy), use that.
     */
    for(;;) {	/* gotoless */
	if((i = toline - fromline) < 0) {
	    if(!len_upcur || (cost = -i * len_upcur) >= gotocost) break;
	    do strcpy(cp, upcur), cp += len_upcur; while(++i);
	} else if((cost = 2 * i)) {	/* lf is mapped to crlf on output */
	    if(cost >= gotocost) break;
	    do *cp++ = '\n'; while(--i);
	    fromcol = 0;
	}
	if((i = tocol - fromcol) < 0) {
	    dist = -i * len_leftcur;
	    if(dist <= len_begoln + tocol) {
		if((cost += dist) > gotocost) break;
		do strcpy(cp, leftcur), cp += len_leftcur; while(++i);
	    } else {
		if((cost += len_begoln) > gotocost) break;
		strcpy(cp, begoln);  cp += len_begoln;
		fromcol = 0;  i = tocol;
	    }
	}
	if(i) {
	    /* if hiliting in effect, just use gotoxy */
	    if(cost + i > gotocost || edattr) break;
	    /* FEATURE: no check is performed whether promptstr consists of
	      printable chars only. gotoxy might cost less if it doesn't */
	    if(fromcol < col0 && toline == line0) {
		strcpy(cp, promptstr);
		cp += strlen(cp);
		fromcol = col0;
	    }
	    strncpy(cp, edbuf + (toline - line0) * cols_1 + fromcol - col0,
		    tocol - fromcol);
	    cp += tocol - fromcol;
	}
	*cp = 0;
	fputs(buf, stdout);
	return;
    }
    gotoxy(tocol, toline);
}

/*
 * clear current input line (but not the prompt)
 * cursor is left right after the prompt
 */
void clearinpline()
{
    if(edlen) {
	int from_c = CURCOL(pos);
	int from_l = CURLINE(pos);
	if(line0 < lines_1) {
	    optmove(from_l, from_c, line0 + 1, 0);
	    fputs(clreoscr, stdout);
	    from_c = 0;  from_l = line0 + 1;
	}
	optmove(from_l, from_c, line0, col0);
	fputs(clreoln, stdout);
    }
}

/*
 * Redraw the input line and put the cursor at the current position.
 * The cursor is assumed to be directly after the prompt.
 */
void redraw_input()
{
    if(edlen) {
	int n, k = cols_1 - col0;
	if(!edattr && *edattrbeg) {
	    edattr = 1;  fputs(edattrbeg, stdout);
	}
	printf("%.*s", k, edbuf);
	if(edlen >= k) {
	    n = lines_1;
	    do {
		printf("\n%.*s", cols_1, edbuf + k);
		--n;  k += cols_1;
	    } while(k <= edlen);
	    if(n < line0) line0 = n;
	}
	optmove(CURLINE(edlen), CURCOL(edlen), CURLINE(pos), CURCOL(pos));
    }
}

/*
 * delete n characters at current position (the position is unchanged)
 * assert(n <= edlen - pos)
 */
void deletechars(n)
int n;
{
    int r_cost, d_cost = n, p = pos;
    int nl = p - CURCOL(p);	/* next line's starting pos */
    int cl = CURLINE(p);	/* current line */
    for(d_cost = p + n; d_cost < edlen; d_cost++)
	edbuf[d_cost - n] = edbuf[d_cost];
    for(d_cost -= n; d_cost < edlen; d_cost++)
	edbuf[d_cost] = ' ';
    for( ; ; printf("\n"), p = nl, cl++) {
	d_cost = 0;
	/* FEATURE: ought to be "d_cost = n > gotocost ? -gotocost : -n;"
	   since redraw will need to goto back. Of little importance */
	if((r_cost = edlen) > (nl += cols_1))
	    r_cost = nl, d_cost = n + gotocost;
	r_cost -= p;
#ifndef VT100
	/* FEATURE: no clreoln is used (it might cost less in the
	  occasion we delete more than one char). Simplicity */
	if(deletecost && deletecost * n + d_cost < r_cost) {
	    for(d_cost = n; d_cost; d_cost--) fputs(delchar, stdout);
	    if(edlen <= nl) break;
	    gotoxy(cols_1 - n, cl);
	    printf("%.*s", n, edbuf + nl - n - 1);
	} else
#endif
	{
	    printf("%.*s", r_cost, edbuf + p);  p += r_cost;
	    if(edlen <= nl) break;
	}
    }
    edbuf[edlen -= n] = 0;
    switch(pos - p) {
	default: optmove(cl, CURCOL(p), CURLINE(pos), CURCOL(pos)); break;
	case 1: fputs(leftcur, stdout); break;
	case 0: break;
    }
}

/*
 * insert n characters at current position. The position is set to after
 * the inserted characters.
 */
void insertchars(str, n)
char *str; int n;
{
    int r_cost, i_cost = n, p = pos;
    int nl = p - CURCOL(p);	/* next line's starting pos */
    int cl = CURLINE(p);	/* current line */
    while(i_cost) strins(edbuf + p, str[--i_cost]);
    edlen += n;  pos += n;
    do {
	i_cost = n;
	if((r_cost = edlen) > (nl += cols_1))
	    r_cost = nl, i_cost += insertfinish;
	r_cost -= p;
#ifndef VT100
	/* FEATURE: insert mode is used only when one char is inserted
	   (which is probaly true > 95% of the time). Simplicity */
	if(n == 1 && insertcost && insertcost + i_cost < r_cost) {
	    printf("%s%c%s", inschar, edbuf[p++], insstop);
	    if(edlen > nl) {
		gotoxy(cols_1, cl);
		fputs(clreoln, stdout);
	    }
	} else
#endif
	{
	    printf("%.*s", r_cost, edbuf + p);  p += r_cost;
	}
	if(edlen < nl) break;
	printf("\n");  p = nl;
	if(++cl > lines_1) cl = lines_1, line0--;
    } while(edlen > nl);
    if(p != pos) optmove(cl, CURCOL(p), CURLINE(pos), CURCOL(pos));
}

/*
 * window size change signal handler:
 * we assume that the bottom line is always the bottom line
 */
void winchsig()
{
    struct winsize wsiz;
  
    /* if ioctl fails or gives silly values, don't change anything */
    if(ioctl(0, TIOCGWINSZ, &wsiz) == 0 && wsiz.ws_row && wsiz.ws_col) {
	lines = wsiz.ws_row; lines_1 = lines - 1;
	cols = wsiz.ws_col; cols_1 = cols - 1;
	send_term_size();
    }
    /* reinstall signal handler for system V unices */
    signal(SIGWINCH, winchsig);
}
